# -*-coding:utf-8-*-
from jieba import Jieba
import multi
import time
import logging
logger = logging.getLogger(__name__)


text = '''網友瘋傳高市新興分局發給各單位協助慈濟骨髓捐贈者戶籍查訪「交辦單」，但不是說「警務大瘦身、大砍20項業務嗎？」引發基層員警議論紛紛；新興分局表示，交辦時間是17日，在警政署20日宣布簡化警察協辦業務前，未來對於類似服務需求將會視個案狀況審慎處理。
警政署20日公布20項簡化協辦業務，讓警察回歸專業，網路即瘋傳一張高市新興分局指派各警勤區員警，到骨髓捐贈者的戶籍地查訪、通告「交辦單」；網友們還在交辦單上打上「署長一直在減別人的爛攤子、高雄卻也是一直撿爛攤子」嘲諷引發群起議論。
不少網友留言：「這種急難救助，不是應該衛生單位訪查嗎？怎會是基層員警？」，也有人認為這太誇張，說「慈濟是收費的，不是免費幫人看病，這不是為民服務，根本是幫慈濟跑腿」；連慈濟義工都看不下，直言「這是我們慈濟人工作，不應交辦給警察，還用他們的警勤時間，做我們的事。」
新興分局指出，慈濟骨髓幹細胞中心請各縣警察局代為協尋已配對成功、但失去聯繫的骨髓幹細胞志願捐贈者，屬急難救助業務一環，盼結合警政分布系統，在最短時間內找尋該志願者，轉送通知信函，速與慈濟骨髓幹細胞中心聯絡，以辦理捐贈移植手續。
新興分局表示，尊重基層同仁想法及建議，對此類服務將儘量婉拒，但如遇到有緊急、非迅速找到配對捐贈者無法救命者，考慮人道精神，仍依個案來協助；絕不給基層員警造成任何工作負荷。'''

texts = [text] * 1000

j = Jieba()


def timeit(method):

    def timed(*args, **kw):
        ts = time.time()
        result = method(*args, **kw)
        te = time.time()
        logger.info('%r %f sec' % (method.__name__, te-ts))
        return result

    return timed


@timeit
def single_test():
    con = []
    for text in texts:
        con.append(j.seg(text))


@timeit
def multi_test():
    multi.seg(texts, 8)


if __name__ == '__main__':
    single_test()
    multi_test()



#-*-coding:utf-8-*-
import re

metachr = '^$*+?{}[]\|()'

heads = u'''>ζↁ━இㅇ（┌／∑┐ψ〒ʕ└⊙＜√＝☜ㄟ〠Σ•ತ✧╰(*-/☮Oε┴‷〈ಸ爻<︿乁♡ｂˋ⁄ˇఠ◉ಠ○vｍ#◎⇎Ψmb◔♀ⶸლ\｡￣◢づ‧≧흫…╭o╮ჰ≡癶凸~ヽヾ'''
tails = u'''ↁﾉ「↑̖〠ิ▽♀⁄ˇ◣～⑤╮凵︿・|━）dʔಠ√ತ'+/°3ืಸ;?ξツఠｍ﹏◎"⇎｡≡╩╭oჴஇ☆┌з┐〒↗⊙☮☞•…*.┴:＼︴♡σ♂ｄ○●Z￣b≦♪╬╰癶>ｼ~＃〉３↘┛”；!✧)Ψ-ⶸ=＿ㅇψˊノ◔ლaづ╧흫m╯y凸'''
for i in metachr:
    heads = heads.replace(i, '\\'+i)
    tails = tails.replace(i, '\\'+i)


def ensure_unicode(string):
    if not isinstance(string, unicode):
        try:
            string = string.decode('utf-8')
        except:
            raise UnicodeError('Input should be UTF8 or UNICODE')
    return string

def clear_num(txt):
    num_pat = re.compile(u'''([0-9\-.:/]{2,})''')
    res = num_pat.sub('', txt)
    return res

def post_check(candlst):
    con = []
    for i in candlst:
        while True:
            i_mod = i.strip()
            i_mod = re.sub(u'\([一-龥]{1,}[\)]{0,1}', '', i_mod)
            i_mod = re.sub(u'（[一-龥]{1,}）', '', i_mod)
            if len(set(i_mod)) < 3:
                break
            else:
                if re.search(u'^m:[0-9a-zA-Z一-龥]', i_mod): # m:大大
                    break
                elif re.search(u'^~[0-9a-zA-Z一-龥]', i_mod): #~4次
                    break
                elif i_mod == i:
                    con.append(i_mod)
                    break
                else:
                   i = i_mod 
                   continue
    return con

def exclude_cht(txt):
    pat = re.compile(u'''(?!.*[.a-zA-Z][.a-zA-Z])([一-龥][一-龥]{2,})''')
    while True:
        res = pat.sub('', txt)
        if res == txt:
            return res
        else:
            txt = res
            exclude_cht(txt)

def find_emo(txt, source=False):
    txt = ensure_unicode(txt)
    txt_mod = clear_num(txt)
    txt_mod = exclude_cht(txt_mod)
    emo_pat = re.compile(u'''(?!.*[.a-zA-Z][.a-zA-Z])([%s].{3,})''' % heads)
    res = emo_pat.findall(txt_mod)
    res = post_check(res)
    res = [i for i in res if i in txt]
    if source == True:
        for x in res:
            txt = txt.replace(x, '<span>'+x+'</span>') 
        return (res,txt)
    return res




# -*-coding:utf-8-*-
from os.path import abspath, dirname, join
import urllib2
import re
import time
import json

CUR_PATH = dirname(abspath(__file__))


def craw_emo(save=True):
    base = 'http://facemood.grtimed.com/index.php?view=facemood&tid=%d'
    pat = re.compile(u'''flashvars="str=(.*?)"''')
    emos = []

    cnt1 = 1
    while cnt1 < 200:
        print cnt1
        url = base % cnt1
        raw = urllib2.urlopen(url).read().decode('utf-8')
        res = pat.findall(raw)
        if res == []:
            cnt1 += 1
        else:
            emos += res
            cnt2 = 2
            while True:
                url2 = url + '&page=%d' % cnt2
                print url2
                raw2 = urllib2.urlopen(url2).read().decode('utf-8')
                res2 = pat.findall(raw2)
                if res2 == []:
                    break
                else:
                    emos += res2
                    cnt2 += 1
                    time.sleep(0.5)
            cnt1 += 1
            time.sleep(0.5)
            print '%d emos collected' % len(emos)
    emos = list(set(emos))
    if save:
        with open('emodata.json', 'w') as jf:
            json.dump(emos, jf)
        print 'emoticons saved!'
    else:
        return emos


def load_emo():
    with open(join(CUR_PATH, 'emoall.txt')) as f:
        emos = f.read().decode('utf-8')
    emos = emos.split('\n')
    emos = [i for i in emos if i != '\n']
    # with open(join(CUR_PATH, 'emodata.json')) as jf1, \
    #      open(join(CUR_PATH, 'emo2.json')) as jf2, \
    #      open(join(CUR_PATH, 'emo3.json')) as jf3, \
    #      open(join(CUR_PATH, 'emo4.json')) as jf4:
    #     emos1 = json.load(jf1)
    #     emos2 = json.load(jf2)
    #     emos3 = json.load(jf3)
    #     emos4 = json.load(jf4)
    # emos = emos1 + emos2 + emos3
    emo_ext = [u'= =', u'=.=', u'=_=', u'>///<', u'> <', u'orz',
               u'^ ^', u'XD', u'(′･_･`)', u'-_-||', u'>"<', u'T.T']
    emos += emo_ext
    emos = list(set(emos))
    emos = [i.strip() for i in emos]
    return emos


def ensure_unicode(string):
    if not isinstance(string, unicode):
        try:
            string = string.decode('utf-8')
        except:
            raise UnicodeError('Input should be UTF8 or UNICODE')
    return string

emos = load_emo()


def find_emo(txt, source=False):
    global emos
    txt = ensure_unicode(txt)
    res = [i for i in emos if i in txt]
    if source is True:
        for x in res:
            txt = txt.replace(x, '<span>' + x + '</span>')
        return (res, txt)
    return res

# source
# http://facemood.grtimed.com/ (顏文字卡)   --> emodata.json
# http://news.gamme.com.tw/55689 (宅宅新聞) --> emo2.json
# http://www.ptt.cc/bbs/cksh77th17/M.1158075878.A.9D2.html (PTT) --> emo3.json
# http://j831124g2009.pixnet.net/blog/post/17973045-素材│日本顏文字（可愛）(*☉౪⊙*)
# (痞客邦) --> emo4.json

# problems
# 兩行以上的emoticons怎麼辦？
#　　　　 ▲           ▲             ▲
# ﹏﹏(ㄏ￣▽￣)ㄏ  q(〒□〒)p    ㄟ(￣▽￣ㄟ)﹏﹏  通通變成阿飄



# -*- coding: utf-8 -*-
"""This is a modified version of "jieba" Chinese segmentator.

All credit goes to fxsjy/jieba.
Find more on: https://github.com/fxsjy/jieba
"""
from multiprocessing import Pool, cpu_count
from functools import partial
from os.path import dirname, abspath, join
from string import punctuation
from math import log
import pickle
import json
import re
import logging

import six

from .emodet import find_emo

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)


CUR_PATH = dirname(abspath(__file__))
with open(join(CUR_PATH, 'brill_tagger.pkl'), 'rb') as f:
    BrillTagger = pickle.load(f)


class Hmm:
    """Hidden Markov Model."""

    def __init__(self):
        """Initialize probabilities."""
        self._min_float = -3.14e100
        self._prev_status = {
            'B': ('E', 'S'),
            'M': ('M', 'B'),
            'S': ('S', 'E'),
            'E': ('B', 'M')
        }

        self._prob = dict()
        probs = ['initP', 'tranP', 'emisP']
        for prob in probs:
            with open(join(CUR_PATH, 'prob', prob + '.json')) as jf:
                self._prob[prob] = json.load(jf)

    def _viterbi(self, obs, states, start_p, trans_p, emit_p):
        _v = [{}]  # tabular
        path = {}
        for y in states:  # init
            _v[0][y] = start_p[y] + emit_p[y].get(obs[0], self._min_float)
            path[y] = [y]
        for t in range(1, len(obs)):
            _v.append({})
            newpath = {}
            for y in states:
                em_p = emit_p[y].get(obs[t], self._min_float)
                (prob, state) = max(
                    [
                        (
                            _v[t - 1][y0] +
                            trans_p[y0].get(y, self._min_float) +
                            em_p, y0
                        )
                        for y0 in self._prev_status[y]
                    ]
                )
                _v[t][y] = prob
                newpath[y] = path[state] + [y]
            path = newpath

        (prob, state) = max([(_v[len(obs) - 1][y], y) for y in ('E', 'S')])

        return (prob, path[state])

    def hmm_seg(self, sentence):
        """Segmentate with HMM method."""
        con = []
        prob, pos_list = self._viterbi(
            sentence, ('B', 'M', 'E', 'S'),
            self._prob['initP'],
            self._prob['tranP'],
            self._prob['emisP']
        )
        begin, next = 0, 0
        for i, char in enumerate(sentence):
            pos = pos_list[i]
            if pos == 'B':
                begin = i
            elif pos == 'E':
                con.append(sentence[begin:i + 1])
                next = i + 1
            elif pos == 'S':
                con.append(char)
                next = i + 1
        if next < len(sentence):
            con.append(sentence[next:])
        return con


class Jieba(Hmm):
    """Main class."""

    def __init__(self, load_ptt_dict=False):
        """__init__ method.

        :param: load_ptt_dict: load PTT dictionary (default: False)
        """
        Hmm.__init__(self)
        self._freq = {}
        self._trie = {}
        self._min_freq = None
        self._gw, self._gwr = {}, {}
        self._gw_num = 0

        self._gen_trie()

        if load_ptt_dict:
            logger.debug('loading ptt dictionary')
            with open(join(CUR_PATH, "ptt_encyc.txt")) as tf:
                raw = tf.read()
                guarantee_wlst = filter(None, raw.split('\n'))
            self.add_guaranteed_wordlist(guarantee_wlst)

    def _load_dic(self):
        logger.debug('loading default dictionary')
        with open(join(CUR_PATH, "dict.txt")) as tf:
            raw = tf.read()
            dic = re.split('\r{0,1}\n', raw)
            dic = filter(None, dic)
            # 字典只得有中文

        return dic

    def add_guaranteed_wordlist(self, lst=None):
        """User-defined dictoinary.

        :param: lst: list of words.
        """
        gw_num = self._gw_num
        if gw_num == 0:
            gw_num = 1
        for num, word in enumerate(lst, gw_num):
            self._gw_num = num
            self._gw[word] = '_gw' + str(num) + '@'
            self._gwr['_gw' + str(num) + '@'] = word

    def _gen_trie(self):
        dic = self._load_dic()
        total = 0.0
        for item in dic:
            word, freq = item.split(' ')
            freq = float(freq)
            self._freq[word] = freq
            total += freq
            p = self._trie
            for char in word:
                if char not in p:
                    p[char] = {}
                p = p[char]
            p[''] = ''  # ending flag

        self._freq = dict(
            [(w, log(float(c) / total)) for w, c in self._freq.items()]
        )
        self._min_freq = min(self._freq.values())

    def _get_dag(self, sentence):
        senlen = len(sentence)
        i, j = 0, 0
        p = self._trie
        dag = {}
        while i < senlen:
            c = sentence[j]
            if c in p:
                p = p[c]
                if '' in p:
                    if i not in dag:
                        dag[i] = []
                    dag[i].append(j)
                j += 1
                if j >= senlen:
                    i += 1
                    j = i
                    p = self._trie
            else:
                p = self._trie
                i += 1
                j = i
        for i in range(len(sentence)):
            if i not in dag:
                dag[i] = [i]
        return dag

    def _calc_route(self, sentence, dag, idx):
        route = {}
        senlen = len(sentence)
        route[senlen] = (0.0, '')
        for idx in range(senlen - 1, -1, -1):
            candidates = [
                (
                    self._freq.get(
                        sentence[idx:x + 1], self._min_freq
                    ) + route[x + 1][0],
                    x
                )
                for x
                in dag[idx]
            ]
            route[idx] = max(candidates)
        return route

    def _seg(self, sentence):  # cut DAG or hmm seg
        dag = self._get_dag(sentence)
        route = self._calc_route(sentence, dag, 0)
        dag_con = []
        x = 0
        buf = u''
        _n = len(sentence)
        while x < _n:
            y = route[x][1] + 1
            l_word = sentence[x:y]
            if y - x == 1:
                buf += l_word
            else:
                if len(buf) > 0:
                    if len(buf) == 1:
                        dag_con.append(buf)
                        buf = u''
                    else:
                        if buf not in self._freq:
                            regognized = self.hmm_seg(buf)
                            for t in regognized:
                                dag_con.append(t)
                        else:
                            for elem in buf:
                                dag_con.append(elem)
                        buf = u''
                dag_con.append(l_word)
            x = y

        if len(buf) > 0:
            if len(buf) == 1:
                dag_con.append(buf)
            else:
                if buf not in self._freq:
                    regognized = self.hmm_seg(buf)
                    for t in regognized:
                        dag_con.append(t)
                else:
                    for elem in buf:
                        dag_con.append(elem)
        return dag_con

    def seg(self, text, pos=False):
        """Segmentate words.

        :param: text: text to be segmentated
        :param: pos: show POS tagging (default: False)
        """
        # find emoticons
        emos = find_emo(text)
        emocan, emocan_r = {}, {}

        for num, emo in enumerate(emos):
            emocan[emo] = '_emo' + str(num) + '@'
            emocan_r['_emo' + str(num) + '@'] = emo

        for emo in emocan.keys():
            sentence = text.replace(emo, emocan[emo])

        # guarantee wlst
        # for gw in self._gw.keys():
        gws = self._gw.keys()
        gws = sorted(gws, key=lambda x: len(x), reverse=True)  # 長詞優先
        for gw in gws:
            if gw in sentence:
                sentence = sentence.replace(gw, self._gw[gw])

        re_han = re.compile(
            r"(http://.*?\s|_gw\d+?@|_emo\d+?@|[ㄅ-ㄩ一-龥]+)", re.U
        )
        # re_ch = re.compile(ur"[ㄅ-ㄩ\u4E00-\u9FA5]+")
        # re_alphnumeric = re.compile(ur'([a-zA-Z0-9]+)')
        re_eng = re.compile(r'[a-zA-Z]+')
        re_neu = re.compile(r'[0-9]+')
        re_url = re.compile(r'(http://.*?)[\s\n\r\n]')
        re_skip = re.compile(r"(\r{0,1}\n|\s)", re.U)

        puns = punctuation
        puns += ''.join([six.unichr(ord(i) + 65248)
                         for i in puns])  # full-width punctuations
        puns += u'、。「」…“”'  # 要在加上一些常用標點符號
        br = u'\r{0,1}\n|\s'
        # metachr = u'^$*+?{}[]\|()'
        re_pun = re.compile(u'(%s|[%s])' % (br, puns))

        blocks = re_han.split(text)

        con = []
        for blk in blocks:
            if blk.startswith('_gw'):
                word = self._gwr[blk]
                logger.debug('find in dict: %s' % word)
                con.append((word, None))

            elif blk.startswith('_emo'):
                word = emocan_r[blk]
                con.append((word, 'EMOTICON'))
            elif re_url.match(blk):
                con.append((re_url.match(blk).group(1), 'URL'))
            # elif re_eng.match(blk):
                # con.append((blk, 'FW'))
            elif re_han.match(blk):
                for word in self._seg(blk):
                    con.append((word, None))
            else:
                tmp = re_pun.split(blk)
                for x in tmp:
                    if x != ' ' and x != '':
                        if re_skip.match(x):
                            con.append(('\n', 'LINEBREAK'))
                        else:
                            con.append((x, None))

        con_w, con_p1 = zip(*con)
        if pos:
            res = BrillTagger.tag(con_w)

            res_con = []
            for word_fin, pos_fin in res:
                if pos_fin == '-None-':
                    if re_pun.match(word_fin):
                        pos_fin = 'PUNCTUATION'
                    elif re_eng.match(word_fin):
                        pos_fin = 'FW'
                    elif re_neu.match(word_fin):
                        pos_fin = 'Neu'
                res_con.append((word_fin, pos_fin))

            word_seg = [i[0] for i in res_con]
            con_p2 = [i[1] for i in res_con]

            pos_con = []
            for pos_p1, pos_p2 in zip(con_p1, con_p2):
                if pos_p1 is None:
                    pos_out = pos_p2
                else:
                    pos_out = pos_p1
                pos_con.append(pos_out)

            output = zip(word_seg, pos_con)
            output_mod = []
            for word, pos in output:
                output_mod.append((word, pos))
            return output_mod
        else:
            return con_w

    def _seg_multi(self, texts, pos=False, pool_size=None):
        """Use `multiprocessing` to segmentate.

        Notice that this is a experimental method, which seems
        to work better with Python3.3+
        :param: texts: list of text
        :param: pos: show POS tags (default: False)
        :param: pool_size: pool size of `multiprocessing`
        """
        import sys
        assert (sys.version_info.major == 3 and sys.version_info.minor >= 3) 
        with Pool(pool_size or cpu_count()) as pool:
            output = pool.map(partial(self.seg, pos=pos), texts)

        return output


# TODO
# add construction detection
# 標點符號分割？
# backslash 可能有問題



# -*-coding:utf-8-*-
from os.path import abspath, dirname, join

CUR_PATH = dirname(abspath(__file__))


def load_emo():
    with open(join(CUR_PATH, 'emoall.txt')) as f:
        emos = f.read()
    emos = emos.split('\n')
    emos = [i for i in emos if i != '\n']
    emo_ext = [u'= =', u'=.=', u'=_=', u'>///<', u'> <', u'orz', u'^ ^', u'XD', u'(′･_･`)', u'-_-||', u'>"<', u'T.T']
    emos += emo_ext
    emos = list(set(emos))
    emos = [i.strip() for i in emos]
    return emos


def find_emo(txt):
    emos = load_emo()
    res = [i for i in emos if i in txt]
    return res



"""Python setup settings."""
from setuptools import setup, find_packages

setup(
    name='jseg',
    version='0.0.2.1',
    packages=find_packages(),
    install_requires=[
        'nltk==3.0.3',
        'six==1.11.0',
    ],
    author='Aji Liu',
    author_email='amigcamel@gmail.com',
    description='''A modified version of Jieba''',
    package_data={
        'jseg': [
            'prob/*',
            'dict.txt',
            'emoall.txt',
            'brill_tagger.pkl'
        ]
    }
)
